/***
 * Juliac
 * Copyright (C) 2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.ultracomp;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.JuliacModuleItf;
import org.objectweb.fractal.juliac.conf.Constants;
import org.objectweb.fractal.juliac.conf.InterceptorSourceCodeGeneratorMap;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.AttributeDesc;
import org.objectweb.fractal.juliac.desc.BindingDesc;
import org.objectweb.fractal.juliac.desc.BindingType;
import org.objectweb.fractal.juliac.desc.ComponentDesc;
import org.objectweb.fractal.juliac.opt.InitializerClassGenerator;
import org.objectweb.fractal.juliac.opt.comp.CompBasedMembraneDesc;
import org.objectweb.fractal.juliac.opt.comp.FCCompCtrlSourceCodeGenerator;
import org.objectweb.fractal.juliac.proxy.ProxyClassGeneratorItf;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.runtime.ultracomp.UltraCompUtils;

/**
 * This class generates the source code associated to Fractal components.
 * The membrane implementation and the initializer implementation are generated.
 * 
 * The content, the interceptors and the controllers are kept in separate
 * classes and the controllers and the interceptors are implemented with
 * components.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.5
 */
public class FCUltraCompCtrlSourceCodeGenerator
extends FCCompCtrlSourceCodeGenerator {
	
	private static final String DEFAULT_CONFIGURATION = "ultracomp-bundled.cfg";
	
	/**
	 * Name of the ADL attribute that defines the type generator for interceptor
	 * components.
	 */
	final private static String COMP_INTERCEPT_TYPE_GEN_ATTRIBUTE_NAME =
		"componentInterceptors";
	
	/**
     * The prefix for symmetrical interfaces of interceptor components.
     * Interceptor components own both an input and an output interface: one
     * corresponds to the intercepted interface type, and the other one is the
     * so-called symmetrical (client <-> server.) Use a prefix instead of a
     * suffix since a suffix introduces name clashes for collection interfaces.
     */
    final static String SYM_ITF_PREFIX = "sym-";
    

    // -----------------------------------------------------------------------
    // Implementation of the JuliacModuleItf interface
    // -----------------------------------------------------------------------

    @Override
    public void init( Juliac jc ) throws IOException {
    	
    	super.init(jc);
    	
        // Redefine the lifecycle interceptor source code generator
    	JuliacConfig jconf = jc.getJuliacConfig();
    	InterceptorSourceCodeGeneratorMap iscgs =
			jconf.interceptorSourceCodeGenerators();
    	iscgs.put(
            Constants.JULIA_LIFECYCLE_CODE_GENERATOR,
            new LifeCycleSourceCodeGenerator() );
    }

    @Override
    protected void postInitJuliaLoaderModule() throws IOException {
        super.postInitJuliaLoaderModule();
        jloader.loadConfigFile(DEFAULT_CONFIGURATION);
    }

    
    // -----------------------------------------------------------------------
    // Implementation of the FCSourceCodeGeneratorItf interface
    // -----------------------------------------------------------------------

    /**
     * Generate the implementation of the membrane associated to a Fractal
     * component.
     * 
     * @param ct                 the component type
     * @param ctrldesc           the component controller descriptor
     * @param contentClassName   the name of the component content class
     * @return                   data about the membrane implementation
     */
    @Override
    public CompBasedMembraneDesc generateMembraneImpl(
        ComponentType ct, String ctrldesc, String contentClassName )
    throws IOException {
    	
    	/*
    	 * TODO the overriden super method generates factories for default
    	 * component-based membranes such as primitive, composite, etc.
    	 * Here the membranes are customized and their factories are embedded in
    	 * the initializer classes. Find a way to reuse these default factories.
    	 * We would need to specialize ComponentDesc in order to introduce a
    	 * special node that simply delegates to a specified definition and
    	 * nothing else.
    	 */
    	    	
    	CompBasedMembraneDesc md =
			super.generateMembraneImpl(ct,ctrldesc,contentClassName);
    	
        InterfaceType[] its = ct.getFcInterfaceTypes();
        String mddescriptor = md.getDescriptor();
        InterfaceType[] mdits = md.getCtrlItfTypes();
        ComponentDesc mdcdesc = md.getCtrlDescs();

        /*
         * Create a mComposite to host the component-based control membrane, the
         * interceptor components and the content. Generate the code and create
         * a descriptor for the mComposite: assign a name that corresponds to
         * the content class name with a heading slash. The heading slash will
         * be used as a naming convention at runtime when traversing the
         * component hierarchy, e.g. from the component control component, to
         * retrieve the mComposite. 
         */
        InterfaceType[] newits =
    		new InterfaceType[ mdits.length + its.length ];
        System.arraycopy(mdits,0,newits,0,mdits.length);
        System.arraycopy(its,0,newits,mdits.length,its.length);
		ComponentType newct = Utils.newBasicComponentType(newits);
		jc.generate(newct,"mComposite",null);
		ComponentDesc cdesc =
			new ComponentDesc(
				"newC0", UltraCompUtils.TOP_LEVEL_PREFIX+contentClassName,
				null, newct, "mComposite", null );
		
		/*
		 * Insert the component-based control membrane into the mComposite.
		 */
		cdesc.addSubComponent(mdcdesc);
		mdcdesc.addSuperComponent(cdesc);
		
		/*
		 * Iterate on the control interface types. Check whether an interceptor
		 * component is needed. Create the needed binding: root -> membrane or
		 * root -> interceptor -> membrane.
		 */
        for( InterfaceType mdit : mdits ) {
    		
        	String mditname = mdit.getFcItfName();
        	ComponentDesc iccdesc = generateInterceptorComponent(cdesc,md,mdit);
        	
        	if( iccdesc == null ) {
        		// No interceptor component
    			if( mdit.isFcClientItf() ) {
    				// Import the client control interface from the membrane
    				BindingDesc bdesc =
    					new BindingDesc(
    						BindingType.IMPORT,mdcdesc,mditname,cdesc,mditname);
    				mdcdesc.putBinding(mditname,bdesc);
    			}
    			else {
    				// Export the server control interface from the membrane
    				BindingDesc bdesc =
    					new BindingDesc(
    						BindingType.EXPORT,cdesc,mditname,mdcdesc,mditname);
    				cdesc.putBinding(mditname,bdesc);
    			}
        	}
    		else {
        		// Interceptor component created.
        		// Create bindings between the component-based control membrane,
        		// the interceptor component and the mComposite.
    			if( mdit.isFcClientItf() ) {

    				// membrane -> interceptor
    				BindingDesc bdesc =
    					new BindingDesc(
    						BindingType.NORMAL,mdcdesc,mditname,iccdesc,mditname);
    				mdcdesc.putBinding(mditname,bdesc);
    					
    				// interceptor -> mComposite
    				BindingDesc bdesc2 =
    					new BindingDesc(
    						BindingType.IMPORT,iccdesc,SYM_ITF_PREFIX+mditname,
    						cdesc,mditname);
    				iccdesc.putBinding(SYM_ITF_PREFIX+mditname,bdesc2);
    			}
    			else {
    				
    				// mComposite -> interceptor
    				BindingDesc bdesc =
    					new BindingDesc(
    						BindingType.EXPORT,cdesc,mditname,iccdesc,mditname);
    				cdesc.putBinding(mditname,bdesc);
    				
    				// interceptor -> membrane
    				BindingDesc bdesc2 =
    					new BindingDesc(
    						BindingType.NORMAL,iccdesc,SYM_ITF_PREFIX+mditname,
    						mdcdesc,mditname);
    				cdesc.putBinding(SYM_ITF_PREFIX+mditname,bdesc2);
    			}
        	}
		}
        
        /*
         * Create a mPrimivite for the content instance.
         * Insert it in the mComposite.
         */
		ComponentDesc contentcdesc = null;
		if( contentClassName != null ) {
			
			// Compute the type of the content mPrimitive component
			List<InterfaceType> l = new ArrayList<InterfaceType>();
			l.addAll(Arrays.asList(its));
			UnifiedClass uc = jc.create(contentClassName);
			
			// BindingController
			UnifiedClass ucbc = jc.create(BindingController.class.getName());
			if( ucbc.isAssignableFrom(uc) ) {
				InterfaceType it =
					new BasicInterfaceType(
						"//content-bc", BindingController.class.getName(),
						false, false, false );
				l.add(it);
			}
			
			// LifeCycleController
			UnifiedClass uclc = jc.create(LifeCycleController.class.getName());
			if( uclc.isAssignableFrom(uc) ) {
				InterfaceType it =
					new BasicInterfaceType(
						"//content-lc", LifeCycleController.class.getName(),
						false, false, false );
				l.add(it);
			}
			
			InterfaceType[] contentits = l.toArray(new InterfaceType[l.size()]);
			ComponentType contentct = Utils.newBasicComponentType(contentits);
			
			// Generate the code for the content mPrimitive component
			jc.generate(contentct,"mPrimitive",contentClassName);
			
			// Insert the content mPrimitive component into the mComposite
			contentcdesc =
				new ComponentDesc(
					UltraCompUtils.CONTENT_COMPONENT_NAME,
					UltraCompUtils.CONTENT_COMPONENT_NAME, null,
					contentct, "mPrimitive", contentClassName );
			cdesc.addSubComponent(contentcdesc);
			contentcdesc.addSuperComponent(cdesc);
			
			// Binding membrane -> content for Binding
			if( ucbc.isAssignableFrom(uc) ) {
				BindingDesc bdesc =
					new BindingDesc(
						BindingType.NORMAL,mdcdesc,"//content-bc",
						contentcdesc,"//content-bc");
				mdcdesc.putBinding("//content-bc",bdesc);				
			}

			// Binding membrane -> content for LifeCycle
			if( uclc.isAssignableFrom(uc) ) {
				BindingDesc bdesc =
					new BindingDesc(
						BindingType.NORMAL,mdcdesc,"//content-lc",
						contentcdesc,"//content-lc");
				mdcdesc.putBinding("//content-lc",bdesc);				
			}
		}
		
		/*
		 * Iterate on the business interface types. Check whether an interceptor
		 * component is needed. Create the needed binding: root -> content or
		 * root -> interceptor -> content.
		 */
        for( InterfaceType it : its ) {
    		
        	String itname = it.getFcItfName();
        	ComponentDesc iccdesc = generateInterceptorComponent(cdesc,md,it);
        	
        	if( iccdesc == null ) {
        		// No interceptor component
        		if( it.isFcClientItf() ) {
    				if( contentClassName == null ) {
    					// no interceptor, no content
    				}
    				else {
    					// content -> mComposite
    					// the binding will be created when the component is bound
    				}        			
        		}
        		else {
    				if( contentClassName == null ) {
    					// no interceptor, no content
    				}
    				else {
    					// mComposite -> content
    					BindingDesc bdesc =
							new BindingDesc(
								BindingType.EXPORT,cdesc,itname,
								contentcdesc,itname);
    					cdesc.putBinding(itname,bdesc);				
    				}        			
        		}
        	}
        	else {
        		// Interceptor component created
    			if( it.isFcClientItf() ) {
    				if( contentClassName != null ) {
    					// content -> interceptor
    					// the binding will be created when the component is bound
    				}
    				// interceptor -> mComposite
    				BindingDesc bdesc2 =
    					new BindingDesc(
    						BindingType.IMPORT,iccdesc,SYM_ITF_PREFIX+itname,
    						cdesc,itname);
    				iccdesc.putBinding(SYM_ITF_PREFIX+itname,bdesc2);
    			}
    			else {    				
    				// mComposite -> interceptor
    				BindingDesc bdesc =
    					new BindingDesc(
    						BindingType.EXPORT,cdesc,itname,iccdesc,itname);
    				cdesc.putBinding(itname,bdesc);
    				if( contentClassName != null ) {
    					// interceptor -> content
    					BindingDesc bdesc2 =
							new BindingDesc(
								BindingType.NORMAL,iccdesc,
								SYM_ITF_PREFIX+itname,contentcdesc,itname);
    					iccdesc.putBinding(SYM_ITF_PREFIX+itname,bdesc2);				
    				}
    			}
        	}			
		}
		
        /*
         * Return a membrane descriptor for the mComposite that contains the
         * component-based control membrane and the interceptor components.
         */
        String mdname = md.getName();
        CompBasedMembraneDesc newmd =
    		new CompBasedMembraneDesc(jc,mdname,mddescriptor,mdits,cdesc);
        
    	return newmd;
    }

    /**
     * Return the source code generator for the initializer class associated
     * with this component source code generator.
     */
    @Override
    public InitializerClassGenerator<ComponentDesc>
    getInitializerClassGenerator() {
        return new InitializerUltraCompCtrlClassGenerator();
    }

    
    // -----------------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------------

    /**
     * If an interceptor component has been created, create it, insert and bind
     * it, and return its descriptor. Else return <code>null</code>.
     * 
     * @param cdesc  the mComposite that hosts the component-based membrane, the
     * 				 interceptor components, and the content
     * @param md     the descriptor of the component-based membrane
     * @param it     the intercepted interface type
     */
    private ComponentDesc generateInterceptorComponent(
		ComponentDesc cdesc, CompBasedMembraneDesc md, InterfaceType it )
    throws IOException {
    	
    	ComponentDesc mdcdesc = md.getCtrlDescs();
    	InterfaceType[] mdits = md.getCtrlItfTypes();
		InterceptorComponentTypeGeneratorItf ictg =
			getInterceptorComponentTypeGenerator(md);

		String itname = it.getFcItfName();
		InterfaceType it0 =
			itname.startsWith("//") ? Utils.downToLevel0InterfaceType(it) : it;
		InterfaceType[] icits = ictg.match(it0,md);
		
		if( icits == null ) {
			//No interceptor component
			return null;
		}
		
		/*
		 * Interceptor component.
		 */
		
		// Retrieve the name of the implementation class for the 
		// interceptor component
		ProxyClassGeneratorItf icg = md.getInterceptorClassGenerator();
		if( icg == null ) {
	        String mddescriptor = md.getDescriptor();
			String msg =
				"Unexpected null interceptor class generator. "+
				mddescriptor+" "+it;
			throw new IOException(msg);
		}
		icg.setInterfaceType(it0);
		icg.setMembraneDesc(md);  // 'md0' for control interfaces but md is ok
		String impl = icg.getTargetClassName();
		
		// Compute the type of the interceptor component by appending the
		// immutable part returned by ictg.match, the current type and the
		// symmetric current type (client <-> server)
		InterfaceType[] fullicits = new InterfaceType[icits.length+2];
		fullicits[0] = it;
		String itsymname = SYM_ITF_PREFIX+it.getFcItfName();
		fullicits[1] = Utils.newSymetricInterfaceType(it,itsymname);
		System.arraycopy(icits,0,fullicits,2,icits.length);
		ComponentType fullicct = Utils.newBasicComponentType(fullicits);
		
		// Generate the code for the interceptor component
		jc.generate(fullicct,"mPrimitive",impl);
		
		// Build a model of the interceptor component and
		// insert it into the mComposite
		String id =
			UltraCompUtils.INTERCEPTOR_COMPONENT_PREFIX +
			UltraCompUtils.toIdentifier(itname);
		ComponentDesc iccdesc =
			new ComponentDesc(id,id,null,fullicct,"mPrimitive",impl);
		cdesc.addSubComponent(iccdesc);
		iccdesc.addSuperComponent(cdesc);
		
		// Bind each client interface of the interceptor component to the
		// component-based control membrane
		for( InterfaceType icit : icits ) {
			
			// Skip client control interfaces (this should be rare)
			String icitname = icit.getFcItfName();
			if( it.isFcClientItf() ) {
				if( itname.equals(icitname) ) {
					continue;
				}
			}
			
			// Skip the symmetrical interface
			if( icitname.startsWith(SYM_ITF_PREFIX) ) {
				continue;
			}
			
			// Check that the component-based control membrane provides
			// the interface required by the interceptor component
			boolean found = false;
			for( InterfaceType mdit : mdits ) {
				String name = mdit.getFcItfName();
				if( name.equals(icitname) ) {
					found = true;
					break;
				}
			}
			if( ! found ) {
				String msg =
					"No such interface "+icitname+
					" for binding the interceptor component in the "+
					md.getDescriptor()+" membrane";
				throw new IOException(msg);
			}
			
			BindingDesc bdesc =
				new BindingDesc(
					BindingType.NORMAL,iccdesc,icitname,mdcdesc,icitname);
			iccdesc.putBinding(icitname,bdesc);
		}

		return iccdesc;
    }
    
    private InterceptorComponentTypeGeneratorItf
    getInterceptorComponentTypeGenerator( CompBasedMembraneDesc md ) {
    	
        InterceptorComponentTypeGeneratorItf[] subs = null;

        ProxyClassGeneratorItf mdiccg = md.getInterceptorClassGenerator();
        if( mdiccg == null ) {
        	
        	/*
        	 * No interceptor control component. This is the case e.g. for
        	 * composite components.
        	 */
        	subs = new InterceptorComponentTypeGeneratorItf[0];
        }
        else {
        	
        	/*
        	 * The 'componentInterceptors' attribute provides the type generator
        	 * for interceptor components. If there is no such attribute, we
        	 * assume that the type corresponds to lifecycle component
        	 * interceptors.
        	 */
        	ComponentDesc iccdesc = md.getInterceptorControlComponent();
            AttributeDesc adesc =
        		iccdesc.getAttribute(COMP_INTERCEPT_TYPE_GEN_ATTRIBUTE_NAME);
            if( adesc == null ) {
            	subs =
        			new InterceptorComponentTypeGeneratorItf[] {
            			new LifeCycleInterceptorComponentTypeGenerator()
            		};
            }
            else {
            	JuliacConfig jconf = jc.getJuliacConfig();

                String value = adesc.getValue();
            	String[] strs = value.split(" ");
            	subs = new InterceptorComponentTypeGeneratorItf[ strs.length ];
            	for( int i=0 ; i < strs.length ; i++ ) {
					Class<InterceptorComponentTypeGeneratorItf> cl =
						jconf.load(strs[i]);
					subs[i] = jconf.instantiate(cl);
				}
            }
        }
        
        InterceptorComponentTypeGeneratorItf ictg =
    		new CompositeInterceptorComponentTypeGenerator(subs);
        return ictg;
    }

    /**
     * Return the optimization level source code generator for control
     * components.
     */
    @Override
    protected JuliacModuleItf
    getControlComponentOptimizationLevelSourceCodeGenerator() {
    	return new FCmUltraCompCtrlSourceCodeGenerator();
    }
}
